/*
 * Hanashizume no Matsuri for X
 * Copyright (c) 2000 TF <tf@denpa.org>
 * All rights reserverd.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *      This product includes software developed by TF.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: filebrowser.c,v 1.10 2001/04/07 13:26:56 tf Exp $
 */
#include <X11/StringDefs.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/List.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/Scrollbar.h>

#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <pwd.h>
#include <unistd.h>

#include "hana.h"

static void listCallback(Widget, XtPointer, XtPointer);
static Widget GetShell(Widget);
static char *doDirname(FileBrowser *, char *);
static Widget CreateList(Widget, FileBrowser *, const char *, const char *);

static void OKCallback(Widget, XtPointer, XtPointer);
static void CancelCallback(Widget, XtPointer, XtPointer);

String translations_fb =
"#override\n"
"<Key>Return: no-op()\n"
"<Key>KP_Enter: no-op()\n"
"<Key>Linefeed: no-op()\n"
"Ctrl<Key>M: no-op()\n"
"Ctrl<Key>J: no-op()\n";

String accel_ok =
"#override\n"
"<Key>Return: set() notify() unset()\n"
"<Key>KP_Enter: set() notify() unset()\n"
"<Key>Linefeed: set() notify() unset()";

void CreateBrowser(Widget parent, FileBrowser *b, const char *name, int type)
{
  Arg args[3];
  int nargs = 0;
  XtAccelerators accel;

  Widget okButton, cancelButton, form, buttonbox;

  b->type = type;

  b->shell = XtCreatePopupShell(name,
			     transientShellWidgetClass, parent,
			     args, nargs);
  XtOverrideTranslations((Widget)b->shell, 
			 XtParseTranslationTable(translations_fb));

  form = XtVaCreateManagedWidget("form",
				 formWidgetClass, b->shell,
				 XtNborderWidth, 0,
				 NULL);
  b->browser = CreateList(form, b, "Open File Named:", "none");

  buttonbox = XtVaCreateManagedWidget("buttonbox",
				      formWidgetClass, form,
				      XtNfromVert, b->browser,
				      XtNright, XtChainRight,
				      XtNleft, XtChainLeft,
				      XtNborderWidth, 0,
				      NULL);

  accel = XtParseAcceleratorTable(accel_ok);
  okButton = XtVaCreateManagedWidget("ok",
				  commandWidgetClass, buttonbox,
				  XtNaccelerators, accel,
				  NULL);
  cancelButton = XtVaCreateManagedWidget("cancel",
					 commandWidgetClass, buttonbox,
					 XtNfromHoriz, okButton,
					 NULL);

  XtAddCallback(okButton, XtNcallback, OKCallback, (XtPointer)b);
  XtAddCallback(cancelButton, XtNcallback, CancelCallback, (XtPointer)b);

  XtInstallAccelerators(b->name, okButton);
}

void OKCallback(Widget w, XtPointer arg, XtPointer junk)
{
  String str;
  char *file, *cp;
  struct stat statbuf;
  FileBrowser *b = (FileBrowser *)arg;

#ifdef XVIDMODE
  if (option_value.screen_mode) {
    EnableScreen();
  }
#endif

  XtVaGetValues(b->name, XtNstring, &str, NULL);
  if (str == NULL || *str == '\0') {
    XawListReturnStruct *lr = XawListShowCurrent(b->list);
    
    if (lr->list_index == XAW_LIST_NONE)
      return;
    str = lr->string;
  }

  /* get a valid string */
  if ((file = doDirname(b, str)) == NULL) {
    fprintf(stderr, "No such file or directory: %s", str);
    return;
  }
  if (stat(file, &statbuf) >= 0 && (statbuf.st_mode & S_IFDIR) != 0) {
    setdir(b, file);
    return;
  }
  if ((cp = strrchr(file, '/')) != NULL) {
    *cp = '\0';
    if (stat(file, &statbuf) >= 0 && (statbuf.st_mode & S_IFDIR) != 0)
      setdir(b, file);
    *cp = '/';
  }

  if (b->type == SAVE_BROWSER) {
    fprintf(stderr, "filebrowser: save to %s\n", file);
    save_game(file);
  } else if (b->type == LOAD_BROWSER) {
    fprintf(stderr, "filebrowser: load from %s\n", file);
    load_game(file);
  } else if (b->type == CAPTURE_BROWSER) {
    fprintf(stderr, "filebrowser: capture image to %s\n", file);
    capture_image(file);
  } else {
    fprintf(stderr, "filebrowser: Unknown type\n");
  }

  XtPopdown(GetShell(w));
}

static void CancelCallback(Widget w, XtPointer arg, XtPointer junk)
{
#ifdef XVIDMODE
  if (option_value.screen_mode) {
    EnableScreen();
  }
#endif
  XtPopdown(GetShell(w));
}

static Widget
CreateList(Widget parent, FileBrowser *b, const char *title_str, const char *dir_name)
{
  Widget form, title;
  static XtTranslations trans;
  char *p = b->type == CAPTURE_BROWSER ? path.imagedir : path.savedir;
  
  trans = XtParseTranslationTable(
"#override\n\
<Key>Return: no-op()\n\
<Key>Linefeed: no-op()\n\
Ctrl<Key>M: no-op()\n\
Ctrl<Key>J: no-op()\n");
  
  form = XtVaCreateManagedWidget("broswer",
				 formWidgetClass, parent,
				 XtNborderWidth, 0,
				 XtNright, XtChainRight,
				 XtNleft, XtChainLeft,
				 XtNtop, XtChainTop,
				 NULL);
  title = XtVaCreateManagedWidget("title",
				  labelWidgetClass, form,
				  XtNborderWidth, 0,
				  XtNlabel, title_str,
				  XtNtop, XtChainTop,
				  XtNbottom, XtChainTop,
				  XtNjustify, XtJustifyLeft,
				  NULL);
  b->name = XtVaCreateManagedWidget("name",
				    asciiTextWidgetClass, form,
				    XtNfromVert, title,
				    XtNeditType, XawtextEdit,
				    XtNtype, XawAsciiString,
				    XtNwrap, XawtextWrapNever,
				    XtNresize, XawtextResizeWidth,
				    XtNtranslations, trans,
				    XtNbottom, XtChainTop,
				    XtNtop, XtChainTop,
				    XtNwidth, 250,
				    NULL);

  b->vport = XtVaCreateManagedWidget("vport",
				     viewportWidgetClass, form,
				     XtNfromVert, b->name,
				     XtNtop, XtChainTop,
				     XtNbottom, XtChainBottom,
				     XtNallowVert, True,
				     XtNallowHoriz, True,
				     XtNuseBottom, True,
				     XtNuseRight, True,
				     XtNwidth, 250,
				     XtNheight, 200,
				     NULL);
  
  b->list = XtVaCreateManagedWidget("files",
				    listWidgetClass, b->vport,
				    XtNborderWidth, 1,
				    XtNverticalList, True,
				    XtNforceColumns, True,
				    XtNdefaultColumns, 1,
				    XtNnumberStrings, 0,
				    NULL);
  
  b->cwd = XtVaCreateManagedWidget("cwd",
				   labelWidgetClass, form,
				   XtNlabel, p,
				   XtNborderWidth, 0,
				   XtNfromVert, b->vport,
				   XtNtop, XtChainBottom,
				   XtNbottom, XtChainBottom,
				   XtNjustify, XtJustifyLeft,
				   XtNresizable, True,
				   NULL);

  b->info = XtVaCreateManagedWidget("info",
				    labelWidgetClass, form,
				    XtNborderWidth, 0,
				    XtNfromVert, b->cwd,
				    XtNtop, XtChainBottom,
				    XtNbottom, XtChainBottom,
				    XtNjustify, XtJustifyLeft,
				    NULL);
  
  XtAddCallback(b->list, XtNcallback, listCallback, (XtPointer)b);

  setdir(b, p);
  
  return form;
}

void setdir(FileBrowser *b, const char *dir)
{
  DIR *dirp;
  struct dirent *e;
  struct stat sb;
  int dirCount = 0, fileCount = 0;
  int count = 0, i = 0;
  char fileStr[MAX_PATH], *filePtr;
  String *slist;
  static char infoStr[256];
  Widget scrollbar;
    
  if (stat(dir, &sb) < 0 || (sb.st_mode & S_IFDIR) == 0) {
    fprintf(stderr, "setdir: Can't open %s\n", dir);
    return;
  }

  if ((dirp = opendir(dir)) == NULL) {
    fprintf(stderr, "setdir: Can't open %s\n", dir);
    return;
  }
  
  while (readdir(dirp) != NULL) count++;
  rewinddir(dirp);

  slist = (String *) XtCalloc(sizeof(String *), count + 1);

  strncpy(fileStr, dir, PATHLEN);
  filePtr = fileStr + strlen(fileStr);
  *filePtr++ = '/';
  
  while ((e = readdir(dirp)) != NULL) {
    struct stat statbuf;
    char *nm = e->d_name;
    
    if (nm[0] == '.') {
      /* skip '.' */
      if (nm[1] == '\0')
	continue;
      if (nm[1] == '.' && nm[2] == '\0') {
	slist[i++] = XtNewString("../ (Go up 1 directory level)");
	continue;
      }
    }
    strncpy(filePtr, nm, PATHLEN);
    if (stat(fileStr, &statbuf) < 0) {
      slist[i++] = XtNewString(nm);
      continue;
    }
    if ((statbuf.st_mode & S_IFDIR) != 0) {
      slist[i] = XtMalloc(sizeof(char) * strlen(nm) + 2);
      strncpy(slist[i], nm, PATHLEN);
      strcat(slist[i], "/");
      i++;
      dirCount++;
      continue;
    }
    /* Now only if it is a file. */
    if ((statbuf.st_mode & (S_IFMT & ~S_IFLNK)) == 0) {
      slist[i++] = XtNewString(nm);
      fileCount++;
    }
  }
  
  closedir(dirp);

  if (dir != NULL) {
    if (b->type == CAPTURE_BROWSER) {
      strncpy(path.imagedir, dir, PATHLEN);
      fprintf(stderr, "image path: %s\n", path.imagedir);
    } else {
      strncpy(path.savedir, dir, PATHLEN);
      fprintf(stderr, "save path: %s\n", path.savedir);
    }
    XtVaSetValues(b->cwd, XtNlabel, dir, NULL);
  }
  XawListChange(b->list, slist, i, 0, True);
  XtVaSetValues(b->name, XtNstring, "", NULL);
  snprintf(infoStr, 256, "%d Directories, %d Files", dirCount, fileCount);
  XtVaSetValues(b->info, XtNlabel, infoStr, NULL);

  if ((scrollbar = XtNameToWidget(b->vport, "vertical")) != None) {
    float top = 0.0;
    
    XawScrollbarSetThumb(scrollbar, top, -1.0);
    XtCallCallbacks(scrollbar, XtNjumpProc, (XtPointer)&top);
  }
}

static void 
listCallback(Widget bar, XtPointer arg, XtPointer itemArg)
{
  XawListReturnStruct *item = (XawListReturnStruct *)itemArg;
  String str, label;
  FileBrowser *b = (FileBrowser *)arg;
  
  if (strncmp(item->string, "../", 3) == 0) {
    label = "../";
  } else {
    label = item->string;
  }
  
  XtVaGetValues(b->name, XtNstring, &str, NULL);
  
  if (strcmp(str, label) == 0) {
    OKCallback(bar, arg, NULL);
  } else {
    XtVaSetValues(b->name, XtNstring, label, NULL);
  }
}

static Widget
GetShell(Widget w)
{
  while (!XtIsShell(w)) {
    w = XtParent(w);
  }
  return w;
}

static char *doDirname(FileBrowser *b, char *file)
{
  static char newPath[MAX_PATH];
  char *cp;
  
  if (file == NULL) {
    if (b->type == CAPTURE_BROWSER)
      strncpy(newPath, path.imagedir, PATHLEN);
    else
      strncpy(newPath, path.savedir, PATHLEN);
    return newPath;
  }
  if (*file == '~') {
    struct passwd *pw;
    
    file++;
    if (*file == '/' || *file == '\0') {
      pw = getpwuid(getuid());
    } else {
      char buf[80], *bp = buf;
      
      while (*file != '/' && *file != '\0')
	*bp++ = *file++;
      *bp = '\0';
      pw = getpwnam(buf);
    }
    if (pw == NULL)
      return NULL;
    while (*file == '/')
      file++;
    strncpy(newPath, pw->pw_dir, PATHLEN);
  } else if (*file == '/') {
    file++;
    newPath[0] = '\0';
  } else {
    if (b->type == CAPTURE_BROWSER)
      strncpy(newPath, path.imagedir, PATHLEN);
    else
      strncpy(newPath, path.savedir, PATHLEN);
  }
  if (strcmp(newPath, "/") != 0) {
    strcat(newPath, "/");
  }
  while (*file != '\0') {
    char *ep;
    
    if ((ep = strchr(file, '/')) == NULL) {
      ep = file + strlen(file);
    }
    if (strncmp(file, "./", 2) == 0 || strcmp(file, ".") == 0) {
      goto bottom;
    }
    if (strncmp(file, "../", 3) == 0 || strcmp(file, "..") == 0) {
      /* First strip trailing '/' */
      char *cp = newPath + strlen(newPath) - 1;
      if (cp != newPath) {
	*cp = '\0';
      }
      cp = strrchr(newPath, '/');
      if (cp == newPath) {
	strncpy(newPath, "/", PATHLEN);
      } else {
	*cp = '\0';
      }
      goto bottom;
    }
    strncat(newPath, file, ep - file);
  bottom:
    file = ep;
    if (*file == '/') {
      strcat(newPath, "/");
    }
    while (*file == '/') {
      file++;
    }
  }
  
  /* Strip any trailing '/'s */
  cp = newPath + strlen(newPath) - 1;
  if (*cp == '/' && cp != newPath) {
    *cp = '\0';
  }
  
  return newPath;
}

void set_filename(FileBrowser *b, const char *filename)
{
  XtVaSetValues(b->name, XtNstring, filename, NULL);
}
