/*
 * Hanashizume no Matsuri for X
 * Copyright (c) 2000, 2001 TF <tf@denpa.org>
 * All rights reserverd.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *      This product includes software developed by TF.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: main.c,v 1.22 2001/04/07 13:26:56 tf Exp $
 */
#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <sys/types.h>
#include <sys/stat.h>

#define MAIN
#include "hana.h"
#ifdef XVIDMODE
#include "vidmode.h"
#endif

#include "Panel.h"

#define APPLICATIONCLASS "XHanashizume"

void quit_proc(Widget, XtPointer, XtPointer);
static void init_system(void);
static void init_path(void);
static void prepare_dir(const char *);
static void RestrictWindowSize(Widget);

String translations =
"<Message>WM_PROTOCOLS: quit()";

String translations_core =
#ifdef XVIDMODE
"Meta<Key>Return: scrwin()\n"
#endif
"Alt<Key>q: quit()\n"
"<Btn3Down>: menupopup()";

static XtResource resources[] = {
#ifdef XVIDMODE
  {"screenMode", "ScreenMode", XtRBool, sizeof(Bool),
   offsetof(OptionValue, screen_mode), XtRImmediate, (XtPointer)False},
#endif
};

static XtActionsRec actions[] = {
  {"quit", (XtActionProc)quit_proc},
  {"menupopup", (XtActionProc)MenuPopup},
#ifdef XVIDMODE
  {"scrwin", ScreenWindow}
#endif
};

static XrmOptionDescRec options[] = {
#ifdef XVIDMODE
  {"-screen", "*screenMode", XrmoptionNoArg, (XtPointer)"True"},
#endif
};

static String fallbacks[] = {
  "*hanacore.font: kanji24",
  "*name.international: False",
  NULL,
};

/* for WM protocol (DELETE) */
Atom wm_protocols[2];

int
main(int argc, char *argv[])
{
  XtSetLanguageProc(NULL, NULL, NULL);

  init_system();
  InitImageProcTable();
  init_path();
  load_global(path.globalinfo);

  top = XtVaOpenApplication(&app, APPLICATIONCLASS,
			    options, XtNumber(options),
			    &argc, argv, fallbacks, sessionShellWidgetClass,
			    NULL);
  XtVaGetApplicationResources(top, (XtPointer)&option_value, resources, XtNumber(resources), NULL);
  XtAppAddActions(app, actions, XtNumber(actions));
  XtOverrideTranslations((Widget)top, XtParseTranslationTable(translations));

  core = XtVaCreateManagedWidget("hanacore", panelWidgetClass, top, NULL);
  XtOverrideTranslations((Widget)core, 
			 XtParseTranslationTable(translations_core));

#ifdef XVIDMODE
  SetupVidMode(top, option_value.screen_mode);
#endif

  XtSetMappedWhenManaged(top, False);
  XtRealizeWidget(top);

  wm_protocols[0] = XInternAtom(XtDisplay(top), "WM_DELETE_WINDOW", False);
  wm_protocols[1] = XInternAtom(XtDisplay(top), "WM_SAVEYOURSELF", False);
  XSetWMProtocols(XtDisplay(top), XtWindow(top), wm_protocols, 2);

  XtMapWidget(top);
  CreateBrowsers(top);
#if XVIDMODE
  if (vm.enable_screen_mode && option_value.screen_mode) {
    XGrabPointer (XtDisplay(top), XtWindow(top), True, 0, GrabModeAsync,
		  GrabModeAsync, XtWindow(top), None, CurrentTime);
  }
#endif    
  RestrictWindowSize(top);

  do {
    XEvent event;
    XtAppNextEvent(app, &event);
    XtDispatchEvent(&event);
  } while (XtAppGetExitFlag(app) == FALSE);

  XtUnrealizeWidget(top);
  XtDestroyWidget(top);

  return 0;
}

void
quit_proc(Widget w, XtPointer dat, XtPointer call)
{
  save_global(path.globalinfo);
  
  XtAppSetExitFlag(XtWidgetToApplicationContext(w));
}

int
open_archives(void)
{
  char file[PATHLEN];

  snprintf(file, PATHLEN, "%s/hanacg.dll", path.orig);
  if ((cg.fd = fopen(file, "rb")) == NULL) {
    fprintf(stderr, "can't open %s\n", file);
    exit(EXIT_FAILURE);
  }
  cg.rlist = create_rlist();
  scan_dll(cg.fd, cg.rlist);
  /* print_rlist(cg.rlist); */

  snprintf(file, PATHLEN, "%s/hanamsg.dll", path.orig);
  if ((msg.fd = fopen(file, "rb")) == NULL) {
    fprintf(stderr, "can't open %s\n", file);
    exit(EXIT_FAILURE);
  }

  msg.rlist = create_rlist();
  scan_dll(msg.fd, msg.rlist);
  /* print_rlist(msg.rlist); */

  return 0;
}

int
close_archives(void)
{
  fclose(cg.fd);
  free_rlist(cg.rlist);

  fclose(msg.fd);
  free_rlist(msg.rlist);
  
  return 0;
}

static void
init_system(void)
{
  sys_stat.speed = SPEED_VFAST;
  sys_stat.skip = False;
  sys_stat.music_mode = MUSIC_NONE;
  sys_stat.music_fd = -1;
  
  sys_stat.music_init[MUSIC_NONE] = False;
  sys_stat.music_init[MUSIC_CDDA] = False;
  sys_stat.music_init[MUSIC_MIDI] = False;
  sys_stat.music_init[MUSIC_MP3] = False;
}

static
void init_path(void)
{
  char basedir[PATHLEN];
  char *p;

  snprintf(basedir, PATHLEN, "%s/%s", getenv("HOME"), ".hanashizume");
  prepare_dir(basedir);

  p = getenv("HANA_ORIGDIR");
  if (p) {
    snprintf(path.orig, PATHLEN, "%s", p);
  } else {
    snprintf(path.orig, PATHLEN, "%s", ORIGDIR);
  }

  p = getenv("HANA_CDDEVICE");
  if (p) {
    snprintf(path.cddevice, PATHLEN, "%s", p);
  } else {
    snprintf(path.cddevice, PATHLEN, "%s", CDDEVICE);
  }

  p = getenv("HANA_MIDIPLAYER");
  if (p) {
    snprintf(path.midiplayer, PATHLEN, "%s", p);
  } else {
    snprintf(path.midiplayer, PATHLEN, "%s", MIDIPLAYER);
  }

  p = getenv("HANA_MP3PLAYER");
  if (p) {
    snprintf(path.mp3player, PATHLEN, "%s", p);
  } else {
    snprintf(path.mp3player, PATHLEN, "%s", MP3PLAYER);
  }

  p = getenv("HANA_MP3DIR");
  if (p) {
    snprintf(path.mp3dir, PATHLEN, "%s", p);
  } else {
    snprintf(path.mp3dir, PATHLEN, "%s", MP3DIR);
  }

  p = getenv("HANA_MP3FILENAME");
  if (p) {
    snprintf(path.mp3filename, PATHLEN, "%s", p);
  } else {
    snprintf(path.mp3filename, PATHLEN, "%s", MP3FILENAME);
  }
  
  snprintf(path.midiwork, PATHLEN, "%s/midi", basedir);
  prepare_dir(path.midiwork);
  snprintf(path.savedir, PATHLEN, "%s/save", basedir);
  prepare_dir(path.savedir);
  snprintf(path.imagedir, PATHLEN, "%s/image", basedir);
  prepare_dir(path.imagedir);

  snprintf(path.globalinfo, PATHLEN, "%s/globalinfo", basedir);
}

static void
prepare_dir(const char *path)
{
  struct stat sb;
  
  if (stat(path, &sb) < 0) {
    fprintf(stderr, "create dir: %s\n", path);
    if (mkdir(path, 0700) < 0) {
      perror(path);
      exit(1);
    }
  } else if (!(sb.st_mode & S_IFDIR)) {
    fprintf(stderr, "%s is not directory.\n", path);
    exit(1);
  }
}

static void
RestrictWindowSize(Widget w)
{
  XSizeHints sh;

  sh.flags = PMinSize | PMaxSize;
  sh.min_width = 640;
  sh.min_height = 480;
  sh.max_width = 640;
  sh.max_height = 480;
  XSetWMNormalHints(XtDisplay(w), XtWindow(w), &sh);
}
