/*
 * Hanashizume no Matsuri for X
 * Copyright (c) 2000 TF <tf@denpa.org>
 * All rights reserverd.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *      This product includes software developed by TF.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: store_game.c,v 1.13 2002/03/22 09:45:40 tf Exp $
 */
#include <stdio.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include "Panel.h"
#include "hana.h"

#ifdef DEBUG
static void print_flags(void);
#endif

int
save_game(const char *filename)
{
  FILE *fp;
  int len;

  if ((fp = fopen(filename, "wb")) == NULL) {
    fprintf(stderr, "save_game: Can't open %s\n", filename);
    return -1;
  }

  /* flag */
  put4b(fp, scn_stat.flag_megumi);
  put4b(fp, scn_stat.flag_yue);
  put4b(fp, scn_stat.flag_ruria);
  put4b(fp, scn_stat.flag_sarasa);
  put4b(fp, scn_stat.flag_hinata);
  put4b(fp, scn_stat.flag_himiko);
  put4b(fp, scn_stat.flag_mao);
  put4b(fp, scn_stat.flag_kurumi);
  put4b(fp, scn_stat.flag_minako);

#ifdef DEBUG
  print_flags();
#endif
  
  /* current state */
  put4b(fp, scn_stat.pe_init_state);
  fprintf(stderr, "init_state: %d\n", scn_stat.pe_init_state);
  put4b(fp, scn_stat.pe_alternative);
  fprintf(stderr, "alternative: %d\n", scn_stat.pe_alternative);

  len = strlen(scn_stat.pe_script) + 1;
  put4b(fp, len);
  fwrite(scn_stat.pe_script, len, 1, fp);
  fprintf(stderr, "script: %s\n", scn_stat.pe_script);
  put4b(fp, scn_stat.pe_block);
  fprintf(stderr, "block: %d\n", scn_stat.pe_block);
  put4b(fp, scn_stat.pe_offset);
  fprintf(stderr, "offset: %x\n", scn_stat.pe_offset);
  
  /* current CG */
  len = strlen(scn_stat.cgname) + 1;
  put4b(fp, len);
  fwrite(scn_stat.cgname, len, 1, fp);

  /* music (0 = stop) */
  put4b(fp, scn_stat.track);

  fclose(fp);

  return 0;
}

int
load_game(const char *filename)
{
  FILE *fp;
  int len;

  StopEngine((PanelWidget)core);

  if ((fp = fopen(filename, "rb")) == NULL) {
    fprintf(stderr, "load_game: Can't open %s\n", filename);
    return -1;
  }

  /* flag */
  scn_stat.flag_megumi = get4b(fp);
  scn_stat.flag_yue = get4b(fp);
  scn_stat.flag_ruria = get4b(fp);
  scn_stat.flag_sarasa = get4b(fp);
  scn_stat.flag_hinata = get4b(fp);
  scn_stat.flag_himiko = get4b(fp);
  scn_stat.flag_mao = get4b(fp);
  scn_stat.flag_kurumi = get4b(fp);
  scn_stat.flag_minako = get4b(fp);

#ifdef DEBUG
  print_flags();
#endif

  /* current state */
  scn_stat.init_state = get4b(fp);
  scn_stat.alternative = get4b(fp);
  fprintf(stderr, "alternative: %d\n", scn_stat.alternative);

  len = get4b(fp);
  fread(scn_stat.script, len, 1, fp);
  fprintf(stderr, "script: %s\n", scn_stat.script);
  scn_stat.block = get4b(fp);
  fprintf(stderr, "block: %d\n", scn_stat.block);
  scn_stat.offset = get4b(fp);
  fprintf(stderr, "offset: %x\n", scn_stat.offset);
  
  /* current CG */
  len = get4b(fp);
  fread(scn_stat.cgname, len, 1, fp);
  fprintf(stderr, "cgname: %s\n", scn_stat.cgname);

  /* music (0 = stop) */
  music_stop();
  scn_stat.track = get4b(fp);
  
  fclose(fp);

  LoadGame((PanelWidget)core);

  return 0;
}

int
put4b(FILE *fp, int val)
{
  unsigned char b1, b2, b3, b4;
  b1 = (val >> 24) & 0xff;
  b2 = (val >> 16) & 0xff;
  b3 = (val >> 8) & 0xff;
  b4 = val & 0xff;

  fputc(b1, fp);
  fputc(b2, fp);
  fputc(b3, fp);
  fputc(b4, fp);

  return 0;
}

int
get4b(FILE *fp)
{
  int b1, b2, b3, b4;
  b1 = fgetc(fp);
  b2 = fgetc(fp);
  b3 = fgetc(fp);
  b4 = fgetc(fp);

  if (b1 == EOF || b2 == EOF || b3 == EOF || b4 == EOF) {
    fprintf(stderr, "save data may be corrupted.\n");
    return 0;
  }

  return (b1 << 24) | (b2 << 16) | (b3 << 8) | b4;
}

#ifdef DEBUG
static void
print_flags(void)
{
  fprintf(stderr, "hinata: %x\n", scn_stat.flag_hinata);
  fprintf(stderr, "kurumi: %x\n", scn_stat.flag_kurumi);
  fprintf(stderr, "sarasa: %x\n", scn_stat.flag_sarasa);
}
#endif
